import os
import unittest

_BENCHMARK = bool(int(os.getenv('TEST_BENCHMARK', default='0')))


@unittest.skipUnless(_BENCHMARK, reason="Benchmark tests are disabled")
class BenchmarkCase(unittest.TestCase):
    pass



import tempfile
from pathlib import Path

from smartparams.io import load, to_string
from tests.unit import UnitCase


class TestToString(UnitCase):
    def test_to_string__yaml_list(self) -> None:
        lst = [None, True, {'1': 2, '3': 4}]

        string = to_string(lst, 'yaml')

        self.assertEqual("- null\n- true\n- '1': 2\n  '3': 4\n", string)

    def test_to_string_yaml_dict(self) -> None:
        dictionary = {'1': [True, False], 3: None}

        string = to_string(dictionary, 'yaml')

        self.assertEqual("'1':\n- true\n- false\n3: null\n", string)

    def test_to_string_yaml_error(self) -> None:
        self.assertRaises(ValueError, to_string, dict(), 'unknown')


class TestLoad(UnitCase):
    def test_load__yaml(self) -> None:
        expected = dict(key='value')
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir, 'config', 'params.yaml')
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_text('key: value\n')

            actual = load(path=path)

            self.assertEqual(expected, actual)

    def test_load__yaml_not_dict(self) -> None:
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir, 'config', 'params.yaml')
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_text('value\n')

            self.assertRaises(ValueError, load, path=path)

    def test_load__unknown(self) -> None:
        path = Path('config', 'params.exe')

        self.assertRaises(ValueError, load, path=path)

    def test_load__no_extension(self) -> None:
        path = Path('config', 'params')

        self.assertRaises(ValueError, load, path=path)



from pathlib import Path
from typing import Any

import yaml


def to_string(data: Any, fmt: str) -> str:
    if fmt in ('yml', 'yaml'):
        return yaml.safe_dump(
            data=data,
            sort_keys=False,
        )

    raise ValueError(f"Format '{fmt}' is not supported.")


def load(path: Path) -> dict[str, Any]:
    if not path.suffix:
        raise ValueError(f"File '{path}' has no extension.")
    elif path.suffix in ('.yml', '.yaml'):
        with path.open() as stream:
            dictionary = yaml.safe_load(stream)
    else:
        raise ValueError(f"File extension '{path.suffix}' is not supported.")

    if isinstance(dictionary, dict):
        return dictionary
    if dictionary is None:
        return dict()
    raise ValueError(f"File '{path}' does not contain a dictionary.")



import inspect
import sys
import warnings
from pathlib import Path
from typing import (
    Any,
    Callable,
    Generic,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
)

import yaml
from typeguard import check_type
from yaml import YAMLError

from smartparams.cli import parse_arguments
from smartparams.io import load, to_string
from smartparams.utils import get_hints, get_name, get_nested, import_class

_T = TypeVar("_T")


class Smart(Generic[_T]):
    name = "Smart"
    keyword = "class"
    missing_value = '???'

    key_separator = '.'
    param_separator = '='
    class_separator = ':'

    default_print_format = 'yaml'

    check_missings: Optional[bool] = None
    check_typings: Optional[bool] = None
    check_overrides: Optional[bool] = None

    _location = '__root__'
    _allow_only_registered = False
    _aliases: dict[str, str] = dict()
    _origins: dict[str, str] = dict()

    def __init__(self, _class: Optional[Type[_T]] = None, /, **params: Any) -> None:
        self._class = _class
        self._params = params

    @property
    def type(self) -> Optional[Type[_T]]:
        return self._class

    @property
    def params(self) -> dict[str, Any]:
        return self._params.copy()

    @classmethod
    def register(
        cls,
        classes: Union[Sequence[str], Mapping[str, str]],
        allow_only_registered: Optional[bool] = None,
    ) -> Type['Smart']:
        if allow_only_registered:
            cls._allow_only_registered = True
        elif allow_only_registered is False and cls._allow_only_registered:
            warnings.warn("Cannot disallow only registered classes if already allowed.")

        if isinstance(classes, Sequence):
            cls._register_classes(classes)
        elif isinstance(classes, Mapping):
            cls._register_aliases(classes)
        else:
            raise TypeError(f"Register classes type '{type(classes)}' is not supported.")

        return cls

    def __call__(self, *args: Any, **kwargs: Any) -> _T:
        if self._class is None:
            raise AttributeError("Class is not set.")

        return self._check_and_init_class(
            location=self._location,
            cls=self._class,
            args=args,
            kwargs=kwargs,
            params=self._instantiate(self._params, self._location),
        )

    def __str__(self) -> str:
        if self._class is None:
            return f"{self.name}({str(self._params)})"

        params = ", ".join((f"{k}={v}" for k, v in self._params.items()))
        return f"{get_name(self._class)}:{self.name}({params})"

    def __repr__(self) -> str:
        return str(self)

    def keys(self) -> list[str]:
        return list(self._params)

    def flatten_keys(self) -> list[str]:
        return self._flatten_keys(self._params)

    def get(self, name: str, init: bool = False, persist: bool = True) -> Any:
        dictionary, key = get_nested(
            dictionary=self._params,
            name=name,
            ensure_key_exists=True,
            separator=self.key_separator,
        )

        if init:
            obj = self._instantiate(dictionary[key], self._location)
            if persist:
                dictionary[key] = obj
            return obj

        return dictionary[key]

    def set(self, name: str, value: Any) -> Any:
        dictionary, key = get_nested(
            dictionary=self._params,
            name=name,
            create_missing_directories=True,
            separator=self.key_separator,
        )
        dictionary[key] = value
        return value

    def pop(self, name: str, init: bool = False) -> Any:
        dictionary, key = get_nested(
            dictionary=self._params,
            name=name,
            ensure_key_exists=True,
            separator=self.key_separator,
        )
        value = dictionary.pop(key)
        if init:
            return self._instantiate(value, self._location)
        return value

    def map(self, name: str, function: Callable) -> Any:
        dictionary, key = get_nested(
            dictionary=self._params,
            name=name,
            ensure_key_exists=True,
            separator=self.key_separator,
        )
        dictionary[key] = value = function(dictionary[key])
        return value

    def update(
        self,
        source: Union['Smart', dict[str, Any], list[str], Path],
        override: bool = True,
    ) -> 'Smart':
        if isinstance(source, Smart):
            self._update_from_smart(source, override)
        elif isinstance(source, dict):
            self._update_from_dict(source, override)
        elif isinstance(source, list):
            self._update_from_list(source, override)
        elif isinstance(source, Path):
            self._update_from_path(source, override)
        else:
            raise TypeError(f"Source type '{type(source)}' is not supported.")

        return self

    def run(
        self,
        function: Optional[Callable] = None,
        path: Optional[Path] = None,
        dump: bool = False,
        save: bool = False,
    ) -> 'Smart':
        args = parse_arguments(
            path=path,
            dump=dump,
            save=save,
        )

        if args.path and args.path.exists():
            self.update(args.path)

        self.update(args.params)

        if args.dump:
            self._dump(
                path=args.path,
                save=args.save,
            )
        elif function is None:
            self()
        else:
            function(self)

        return self

    @classmethod
    def _register_classes(cls, classes: Sequence[str]) -> None:
        cls._register_aliases({item: item for item in classes})

    @classmethod
    def _register_aliases(cls, aliases: Mapping[str, str]) -> None:
        for origin, alias in aliases.items():
            if origin in cls._aliases:
                warnings.warn(f"Override origin '{origin}'.")
                cls._origins.pop(cls._aliases.pop(origin))

            if alias in cls._origins:
                warnings.warn(f"Override alias '{alias}'.")
                cls._aliases.pop(cls._origins.pop(alias))

            cls._aliases[origin] = alias
            cls._origins[alias] = origin

    def _representation(self, obj: Any, skip_default: bool = False) -> dict[str, Any]:
        representation: dict[str, Any] = dict()
        signature = inspect.signature(obj)

        for name, param in signature.parameters.items():
            if name != 'self' and param.kind in (
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.KEYWORD_ONLY,
            ):
                annotation = param.annotation
                default = param.default

                if annotation is Smart or isinstance(default, Smart) and default.type is None:
                    representation[name] = {
                        self.keyword: self.name,
                    }
                elif get_origin(annotation) is Smart or isinstance(default, Smart):
                    if isinstance(default, Smart):
                        param_type = default.type
                    else:
                        param_type, *_ = get_args(annotation)

                    keyword = inspect.formatannotation(param_type)
                    keyword = self._aliases.get(keyword, keyword)
                    keyword = keyword + self.class_separator + self.name

                    representation[name] = {
                        self.keyword: keyword,
                        **self._representation(param_type, skip_default),
                    }
                elif default is not inspect.Parameter.empty and skip_default:
                    continue
                elif default is None or isinstance(default, (bool, float, int, str)):
                    representation[name] = default
                elif annotation is not inspect.Parameter.empty and isinstance(annotation, type):
                    if annotation in (bool, float, int, str):
                        representation[name] = annotation.__name__ + self.missing_value
                    else:
                        keyword = inspect.formatannotation(annotation)
                        keyword = self._aliases.get(keyword, keyword)
                        representation[name] = {
                            self.keyword: keyword,
                            **self._representation(annotation, skip_default),
                        }
                else:
                    representation[name] = self.missing_value

        return representation

    def _flatten_keys(self, obj: Any, prefix: str = "") -> list[str]:
        if not isinstance(obj, dict):
            return [prefix]

        keys = []
        for k, v in obj.items():
            if prefix:
                k = prefix + self.key_separator + k
            keys.extend(self._flatten_keys(v, k))

        return keys

    def _update_from_smart(self, smart: 'Smart', override: bool) -> None:
        flatten_params = {key: self.pop(key) for key in self.flatten_keys()}

        for key in smart.flatten_keys():
            self.set(key, smart.get(key))
            if override:
                flatten_params.pop(key, None)

        for key, value in flatten_params.items():
            self.set(key, value)

    def _update_from_dict(self, params: dict[str, Any], override: bool) -> None:
        self._update_from_smart(Smart(**params), override)

    def _update_from_list(self, params_list: list[str], override: bool) -> None:
        smart: Smart = Smart()
        for param in params_list:
            key, separator, raw_value = param.partition(self.param_separator)

            if not separator:
                raise ValueError(f"Param '{param}' does not contain '=' mark.")

            try:
                value = yaml.safe_load(raw_value)
            except YAMLError as e:
                raise ValueError(f"Param '{param}' has invalid value.") from e

            smart.set(key, value)

        self._update_from_smart(smart, override)

    def _update_from_path(self, path: Path, override: bool) -> None:
        self._update_from_dict(load(path), override)

    def _instantiate(self, obj: Any, _location: str = '') -> Any:
        location = _location or self._location

        if isinstance(obj, dict):
            if self.keyword in obj:
                return self._instantiate_class(obj, location)

            return self._instantiate_dict(obj, location)

        if isinstance(obj, list):
            return self._instantiate_list(obj, location)

        return obj

    def _instantiate_dict(self, dictionary: dict[str, Any], location: str) -> dict[str, Any]:
        return {
            key: self._instantiate(value, f"{location}{self.key_separator}{key}")
            for key, value in dictionary.items()
        }

    def _instantiate_list(self, lst: list[Any], location: str) -> list[Any]:
        return [
            self._instantiate(element, f"{location}{self.key_separator}{index}")
            for index, element in enumerate(lst)
        ]

    def _instantiate_class(self, dictionary: dict[str, Any], location: str) -> Any:
        kwargs = dictionary.copy()
        class_name = kwargs.pop(self.keyword)
        class_name, _, option = class_name.partition(self.class_separator)

        if class_name == self.name:
            return self._check_and_init_class(
                location=location,
                cls=Smart,
                kwargs=kwargs,
            )

        if class_name in self._origins:
            class_name = self._origins[class_name]
        elif self._allow_only_registered:
            raise ImportError(f"Class '{class_name}' is not registered.")

        cls = cast(Type[_T], import_class(class_name))

        if option:
            if option == self.name:
                return self._check_and_init_class(
                    location=location,
                    cls=Smart,
                    args=(cls,),
                    kwargs=kwargs,
                )
            else:
                raise ValueError(f"Option '{option}' is not supported.")
        else:
            return self._check_and_init_class(
                location=location,
                cls=cls,
                kwargs=kwargs,
            )

    def _check_and_init_class(
        self,
        location: str,
        cls: Type[Any],
        args: Optional[tuple[Any, ...]] = None,
        kwargs: Optional[dict[str, Any]] = None,
        params: Optional[dict[str, Any]] = None,
    ) -> Any:
        location_with_classname = f"{location}{self.class_separator}{get_name(cls)}"
        args = args or tuple()
        kwargs = kwargs or dict()

        if params:
            if self.check_overrides is not False:
                self._check_overrides(
                    params=params,
                    kwargs=kwargs,
                    location=location_with_classname,
                )

            params = params.copy()
            params.update(kwargs)
            kwargs = params

        if self.check_missings is not False:
            self._check_missings(
                kwargs=kwargs,
                location=location_with_classname,
            )

        if self.check_typings is not False:
            self._check_typings(
                cls=cls,
                args=args,
                kwargs=kwargs,
                location=location_with_classname,
            )

        try:
            obj = cls(*args, **kwargs)
        except Exception as e:
            raise RuntimeError(f"Error during instantiate {location_with_classname} class.") from e
        else:
            if isinstance(obj, Smart):
                obj._location = location
            return obj

    def _check_overrides(
        self,
        location: str,
        params: dict[str, Any],
        kwargs: dict[str, Any],
    ) -> None:
        if overrides := set(params).intersection(kwargs):
            msg = f"Override {location}'s class arguments {overrides}."
            if self.check_overrides:
                raise ValueError(msg)
            warnings.warn(msg)

    def _check_missings(
        self,
        location: str,
        kwargs: dict[str, Any],
    ) -> None:
        for name, value in kwargs.items():
            if isinstance(value, str) and value.endswith(self.missing_value):
                msg = f"Missing {location}'s class argument '{name}' value."
                if self.check_missings:
                    raise ValueError(msg)
                warnings.warn(msg)

    def _check_typings(
        self,
        location: str,
        cls: Type[Any],
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> None:
        signature = inspect.signature(cls)
        arguments = signature.bind(*args, **kwargs).arguments

        for name, expected_type in get_hints(signature).items():
            if name in arguments:
                try:
                    check_type(
                        argname=f"{location}'s class argument '{name}'",
                        value=arguments[name],
                        expected_type=expected_type,
                    )
                except TypeError as e:
                    msg = f"The {''.join(e.args)}."
                    if self.check_typings:
                        raise TypeError(msg)
                    warnings.warn(msg)

    def _dump(self, path: Optional[Path], save: bool) -> None:
        self.update(
            self._representation(self._class, skip_default=bool(self._params)),
            override=False,
        )

        print_format = path.suffix.strip('.') if path else self.default_print_format
        text_params = to_string(self.params, print_format)
        text_keys = to_string(self.flatten_keys(), print_format)

        print(f"{text_params}---\n{text_keys}")

        if save:
            if path:
                path.write_text(text_params)
                print(f"Saved params in {path}.")
            else:
                print("Cannot save if path is not specified.", file=sys.stderr)



from pathlib import Path
from unittest.mock import Mock, patch

from smartparams.cli import CLIArguments, _parse_arguments, parse_arguments
from tests.unit import UnitCase


class TestParseArguments(UnitCase):
    @patch('smartparams.cli.sys', Mock(argv=['script.py', 'arg1', 'arg2']))
    @patch('smartparams.cli._parse_arguments')
    def test_parse_arguments(self, _parse_arguments: Mock) -> None:
        parse_arguments(
            path=None,
            dump=True,
            save=True,
        )

        _parse_arguments.assert_called_once_with(
            arguments=['arg1', 'arg2'],
            path=None,
            dump=True,
            save=True,
        )

    def test_parse_arguments__without_cli(self) -> None:
        arguments: list[str] = []
        path = Path('/home/params.yaml')
        dump = True
        save = True
        expected = CLIArguments(
            path=Path('/home/params.yaml'),
            dump=True,
            save=True,
            params=[],
        )

        actual = _parse_arguments(
            arguments=arguments,
            path=path,
            dump=dump,
            save=save,
        )

        self.assertEqual(expected, actual)

    def test_parse_arguments__with_cli(self) -> None:
        arguments = [
            '--path',
            'cli_params.yaml',
            '--dump',
            '--save',
            '-s',
            'key1=value1',
            '--set',
            'key2=value2',
        ]
        path = Path('/home/params.yaml')
        dump = True
        save = True
        expected = CLIArguments(
            path=Path('/home/cli_params.yaml'),
            dump=True,
            save=True,
            params=['key1=value1', 'key2=value2'],
        )

        actual = _parse_arguments(
            arguments=arguments,
            path=path,
            dump=dump,
            save=save,
        )

        self.assertEqual(expected, actual)

    def test_parse_arguments__with_cli_absolute_path(self) -> None:
        arguments = [
            '--path',
            '/cli_params.yaml',
            '--dump',
            '--save',
            '-s',
            'key1=value1',
            '--set',
            'key2=value2',
        ]
        path = Path('/home/params.yaml')
        dump = True
        save = True
        expected = CLIArguments(
            path=Path('/cli_params.yaml'),
            dump=True,
            save=True,
            params=['key1=value1', 'key2=value2'],
        )

        actual = _parse_arguments(
            arguments=arguments,
            path=path,
            dump=dump,
            save=save,
        )

        self.assertEqual(expected, actual)



from copy import deepcopy
from pathlib import Path
from typing import Any, Union
from unittest.mock import Mock, patch

from smartparams.smart import Smart
from tests.custom_classes import Class
from tests.unit import UnitCase


class TestSmartInitCase(UnitCase):
    def setUp(self) -> None:
        self.smart = Smart(Class, arg1='arg1', arg2=10)

    def test_params(self) -> None:
        params = self.smart.params

        self.assertEqual(dict(arg1='arg1', arg2=10), params)
        self.assertIsNot(self.smart._params, params)

    def test_call(self) -> None:
        obj = self.smart()

        self.assertIsInstance(obj, Class)
        self.assertEqual('arg1', obj.arg1)
        self.assertEqual(10, obj.arg2)

    def test_call__with_params(self) -> None:
        smart = Smart(Class)

        obj = smart('a1', arg2=15)

        self.assertIsInstance(obj, Class)
        self.assertEqual('a1', obj.arg1)
        self.assertEqual(15, obj.arg2)

    def test_call__with_duplicated_params(self) -> None:
        smart = Smart(Class, arg1='arg1')

        self.assertRaises(TypeError, smart, 'a1')

    def test_call__without_class(self) -> None:
        smart: Smart = Smart()

        self.assertRaises(AttributeError, smart)


class TestSmartAccessCase(UnitCase):
    def setUp(self) -> None:
        self.smart: Smart = Smart(arg1='arg1', arg2=['arg2'], arg3={'arg31': 'a31', 'arg32': 'a32'})

    def test_keys(self) -> None:
        keys = self.smart.keys()

        self.assertTupleEqual(('arg1', 'arg2', 'arg3'), tuple(keys))

    def test_get(self) -> None:
        value = self.smart.get('arg3.arg31')

        self.assertEqual('a31', value)

    def test_set(self) -> None:
        new_value = 'argument31'

        value = self.smart.set('arg3.arg31', new_value)

        self.assertEqual('argument31', value)
        self.assertEqual('argument31', self.smart.params['arg3']['arg31'])
        self.assertEqual('a32', self.smart.params['arg3']['arg32'])

    def test_pop(self) -> None:
        value = self.smart.pop('arg3.arg31')

        self.assertEqual('a31', value)
        self.assertFalse('arg3' in self.smart.params['arg3'])

    def test_map(self) -> None:
        function = Mock(return_value='argument31')

        value = self.smart.map('arg3.arg31', function)

        self.assertEqual('argument31', value)
        self.assertEqual('argument31', self.smart.params['arg3']['arg31'])
        self.assertEqual('a32', self.smart.params['arg3']['arg32'])

    @patch('smartparams.smart.load')
    def test_update(self, load: Mock) -> None:
        dictionary = {'arg3': {'arg31': 'argument31'}, 'arg4': 'argument4'}
        load.return_value = dictionary
        test_cases: list[tuple[Union['Smart', dict[str, Any], list[str], Path], str]] = [
            (Smart(**dictionary), "smart"),
            (dictionary, "dict"),
            (['arg3.arg31=argument31', 'arg4=argument4'], "list"),
            (Path('path/to/file.yaml'), "path"),
        ]

        for source, msg in test_cases:
            smart: Smart = deepcopy(self.smart)

            with self.subTest(msg=msg):
                smart.update(source)

                self.assertTrue('arg1' in smart.params)
                self.assertTrue('arg2' in smart.params)
                self.assertTrue('arg3' in smart.params)
                self.assertTrue('arg4' in smart.params)
                self.assertEqual('arg1', smart.params['arg1'])
                self.assertListEqual(['arg2'], smart.params['arg2'])
                self.assertEqual('argument31', smart.params['arg3']['arg31'])
                self.assertEqual('a32', smart.params['arg3']['arg32'])
                self.assertEqual('argument4', smart.params['arg4'])

    def test_update__unknown(self) -> None:
        source = tuple()  # type: ignore

        self.assertRaises(TypeError, self.smart.update, source)

    def test_flatten_keys(self) -> None:
        value = self.smart.flatten_keys()

        self.assertListEqual(['arg1', 'arg2', 'arg3.arg31', 'arg3.arg32'], value)


class TestSmartInstantiateCase(UnitCase):
    def setUp(self) -> None:
        self.class_name = f"{Class.__module__}.{Class.__qualname__}"
        self.params = dict(
            smart_dict={'class': 'Smart'},
            smart={'class': self.class_name + ':Smart', 'arg1': 'arg1', 'arg2': 10},
            object={'class': self.class_name, 'arg1': 'arg1', 'arg2': 10},
            value=21,
        )
        self.smart: Smart = Smart(**self.params)

    def test_get__init_smart(self) -> None:
        obj = self.smart.get('smart', init=True)

        self.assertIsInstance(obj, Smart)
        self.assertIs(obj.type, Class)
        self.assertTrue('smart' in self.smart.params)

    def test_get__init_object(self) -> None:
        obj = self.smart.get('object', init=True)

        self.assertIsInstance(obj, Class)
        self.assertTrue('object' in self.smart.params)

    def test_get__init_value(self) -> None:
        obj = self.smart.get('value', init=True)

        self.assertIsInstance(obj, int)
        self.assertTrue('value' in self.smart.params)

    def test_get__init_persist(self) -> None:
        obj = self.smart.get('object', init=True, persist=True)

        self.assertIs(self.smart.get('object', init=True, persist=False), obj)
        self.assertIsInstance(self.smart.get('object'), Class)
        self.assertTrue('object' in self.smart.params)

    def test_get__init_not_persist(self) -> None:
        obj = self.smart.get('object', persist=False)

        self.assertIsNot(self.smart.get('object', init=True, persist=False), obj)
        self.assertIsInstance(self.smart.get('object'), dict)
        self.assertTrue('object' in self.smart.params)

    def test_pop__init_smart(self) -> None:
        obj = self.smart.pop('smart', init=True)

        self.assertIsInstance(obj, Smart)
        self.assertIs(obj.type, Class)
        self.assertFalse('smart' in self.smart.params)

    def test_pop__init_object(self) -> None:
        obj = self.smart.pop('object', init=True)

        self.assertIsInstance(obj, Class)
        self.assertFalse('object' in self.smart.params)

    def test_pop__init_value(self) -> None:
        obj = self.smart.pop('value', init=True)

        self.assertIsInstance(obj, int)
        self.assertFalse('value' in self.smart.params)

    def test_instantiate__any(self) -> None:
        data = "string"

        result = self.smart._instantiate(data)

        self.assertEqual(data, result)

    def test_instantiate__smart_dict(self) -> None:
        obj = self.params['smart_dict']

        result = self.smart._instantiate(obj)

        self.assertIsInstance(result, Smart)
        self.assertIsNone(result.type)

    def test_instantiate__smart(self) -> None:
        obj = self.params['smart']

        result = self.smart._instantiate(obj)

        self.assertIsInstance(result, Smart)
        self.assertIs(result.type, Class)

    def test_instantiate__unknown(self) -> None:
        obj = self.params['smart']
        obj['class'] += "suffix"  # type: ignore

        self.assertRaises(ValueError, self.smart._instantiate, obj)

    def test_instantiate__object(self) -> None:
        obj = self.params['object']

        result = self.smart._instantiate(obj)

        self.assertIsInstance(result, Class)

    def test_instantiate__dict(self) -> None:
        obj = {'key': 'value'}

        result = self.smart._instantiate(obj)

        self.assertIsNot(obj, result)

    def test_instantiate__list(self) -> None:
        obj = ['item']

        result = self.smart._instantiate(obj)

        self.assertIsNot(obj, result)

    def test_instantiate___with_aliases(self) -> None:
        self.smart._allow_only_registered = True
        self.smart._aliases = {self.class_name: 'Class'}
        self.smart._origins = {v: k for k, v in self.smart._aliases.items()}
        obj = self.params['object']
        obj['class'] = 'Class'  # type: ignore

        result = self.smart._instantiate(obj)

        self.assertIsInstance(result, Class)

    def test_instantiate___with_aliases_not_used(self) -> None:
        self.smart._aliases = {self.class_name: 'Class'}
        self.smart._origins = {v: k for k, v in self.smart._aliases.items()}
        obj = self.params['object']

        result = self.smart._instantiate(obj)

        self.assertIsInstance(result, Class)

    def test_instantiate___with_aliases_error(self) -> None:
        self.smart._allow_only_registered = True
        self.smart._aliases = {self.class_name: 'Class'}
        self.smart._origins = {v: k for k, v in self.smart._aliases.items()}
        obj = self.params['object']

        self.assertRaises(ImportError, self.smart._instantiate, obj)


class TestRegisterCase(UnitCase):
    def setUp(self) -> None:
        self._aliases = Smart._aliases.copy()
        self._origins = Smart._origins.copy()
        self._allow_only_registered = Smart._allow_only_registered

    def tearDown(self) -> None:
        Smart._aliases = self._aliases.copy()
        Smart._origins = self._origins.copy()
        Smart._allow_only_registered = self._allow_only_registered

    def test_register__classes(self) -> None:
        classes = ('origin.1', 'origin.2')

        Smart.register(classes)

        self.assertFalse(Smart._allow_only_registered)
        self.assertEqual({'origin.1': 'origin.1', 'origin.2': 'origin.2'}, Smart._aliases)
        self.assertEqual({'origin.1': 'origin.1', 'origin.2': 'origin.2'}, Smart._origins)

    def test_register__aliases(self) -> None:
        aliases = {'origin.1': 'alias.1', 'origin.2': 'alias.2'}

        Smart.register(aliases)

        self.assertFalse(Smart._allow_only_registered)
        self.assertEqual({'origin.1': 'alias.1', 'origin.2': 'alias.2'}, Smart._aliases)
        self.assertEqual({'alias.1': 'origin.1', 'alias.2': 'origin.2'}, Smart._origins)

    @patch('smartparams.smart.warnings')
    def test_register__duplicates(self, warnings: Mock) -> None:
        Smart._aliases = {'origin.1': 'alias.1', 'origin.2': 'alias.2'}
        Smart._origins = {'alias.1': 'origin.1', 'alias.2': 'origin.2'}
        aliases = {'origin.1': 'alias.2'}

        Smart.register(aliases)

        self.assertEqual({'origin.1': 'alias.2'}, Smart._aliases)
        self.assertEqual({'alias.2': 'origin.1'}, Smart._origins)
        warnings.warn.assert_called()

    def test_register__allow_only_registered(self) -> None:
        classes = ('origin.1', 'origin.2')

        Smart.register(classes, allow_only_registered=True)

        self.assertTrue(Smart._allow_only_registered)

    @patch('smartparams.smart.warnings')
    def test_register__disallow_only_registered(self, warnings: Mock) -> None:
        Smart._allow_only_registered = True
        classes = ('origin.1', 'origin.2')

        Smart.register(classes, allow_only_registered=False)

        self.assertTrue(Smart._allow_only_registered)
        warnings.warn.assert_called()

    def test_register__raise(self) -> None:
        self.assertRaises(TypeError, Smart.register, Mock())



import pickle as pkl
import sys
import tempfile
from pathlib import Path
from unittest.mock import Mock, patch

from smartparams import Smart
from tests.custom_classes import (
    Class,
    ClassChild,
    ClassComposition,
    RaiseClass,
    some_function,
)
from tests.unit import UnitCase


class TestSerialization(UnitCase):
    def setUp(self) -> None:
        class_name = f"{Class.__module__}.{Class.__qualname__}"
        class_child_name = f"{ClassChild.__module__}.{ClassChild.__qualname__}"
        self.smart = Smart(
            ClassComposition,
            **{
                'cls': {'arg1': 'argument1', 'arg2': 15, 'class': class_name},
                'smart': {'class': 'Smart'},
                'smart_cls': {'arg1': 'str???', 'arg2': 75, 'class': f'{class_name}:Smart'},
                'unknown': some_function,
                'smart_cls_with_default': {
                    'arg1': 'argument1',
                    'arg2': 5,
                    'class': f'{class_child_name}:Smart',
                },
            },
        )

    def test_pickle(self) -> None:
        pickled = pkl.dumps(self.smart)
        smart = pkl.loads(pickled)

        self.assertIsInstance(smart, Smart)
        self.assertEqual(str(self.smart), str(smart))
        self.assertIs(smart.type, ClassComposition)
        self.assertEqual(self.smart.params, smart.params)


class TestSmartRunCase(UnitCase):
    def setUp(self) -> None:
        self.smart = Smart(Class, arg2=10)

    @patch('smartparams.cli.sys', Mock(argv=[]))
    def test_run__dump_save_with_path(self) -> None:
        with tempfile.NamedTemporaryFile(mode='w+', suffix='.yaml') as file:
            self.smart.run(path=Path(file.name), dump=True, save=True)

            self.assertEqual("arg1: str???\narg2: 10\n", file.read())

    @patch('smartparams.cli.sys', Mock(argv=[]))
    @patch('smartparams.smart.print')
    def test_run__dump_without_path(self, print_mock: Mock) -> None:
        self.smart.run(path=None, dump=True)

        print_mock.assert_called_once_with("arg1: str???\narg2: 10\n---\n- arg1\n- arg2\n")

    @patch('smartparams.cli.sys', Mock(argv=[]))
    @patch('smartparams.smart.print')
    def test_run__dump_save_without_path(self, print_mock: Mock) -> None:
        self.smart.run(path=None, dump=True, save=True)

        print_mock.assert_called_with("Cannot save if path is not specified.", file=sys.stderr)

    @patch('smartparams.smart.load', Mock(return_value={'arg1': 'string'}))
    def test_run__function(self) -> None:
        function = Mock()
        with tempfile.NamedTemporaryFile(mode='w+', suffix='.yaml') as file:
            file.write("arg1: string\n")
            file.seek(0)

            self.smart.run(function=function, path=Path(file.name))

            function.assert_called_once_with(self.smart)

    @patch('smartparams.smart.load', Mock(return_value={'arg1': 'string'}))
    def test_run__self(self) -> None:
        self.smart._class = Mock()  # type: ignore
        with tempfile.NamedTemporaryFile(mode='w+', suffix='.yaml') as file:
            file.write("arg1: string\n")
            file.seek(0)

            self.smart.run(path=Path(file.name))

            self.smart._class.assert_called_once_with(arg1='string', arg2=10)  # type: ignore


class TestSmartRepresentationCase(UnitCase):
    def setUp(self) -> None:
        self.smart: Smart = Smart()

    def test_representation__with_defaults(self) -> None:
        class_name = f"{Class.__module__}.{Class.__qualname__}"
        class_child_name = f"{ClassChild.__module__}.{ClassChild.__qualname__}"
        expected = {
            'cls': {'class': class_name, 'arg1': 'str???', 'arg2': 5},
            'smart_cls': {'class': f'{class_name}:Smart', 'arg1': 'str???', 'arg2': 5},
            'smart': {'class': 'Smart'},
            'unknown': '???',
            'smart_cls_with_default': {
                'class': f'{class_child_name}:Smart',
                'arg1': 'str???',
                'arg2': 5,
            },
            'smart_cls_with_only_default_generic': {
                'class': 'Smart',
            },
        }

        actual = self.smart._representation(ClassComposition, skip_default=False)

        self.assertEqual(expected, actual)

    def test_representation__without_defaults(self) -> None:
        class_name = f"{Class.__module__}.{Class.__qualname__}"
        class_child_name = f"{ClassChild.__module__}.{ClassChild.__qualname__}"
        expected = {
            'cls': {'class': class_name, 'arg1': 'str???'},
            'smart_cls': {'class': f'{class_name}:Smart', 'arg1': 'str???'},
            'smart': {'class': 'Smart'},
            'unknown': '???',
            'smart_cls_with_default': {
                'class': f'{class_child_name}:Smart',
                'arg1': 'str???',
            },
            'smart_cls_with_only_default_generic': {
                'class': 'Smart',
            },
        }

        actual = self.smart._representation(ClassComposition, skip_default=True)

        self.assertEqual(expected, actual)

    def test_representation__with_aliases(self) -> None:
        class_name = f"{Class.__module__}.{Class.__qualname__}"
        class_child_name = f"{ClassChild.__module__}.{ClassChild.__qualname__}"
        self.smart._allow_only_registered = True
        self.smart._aliases = {class_name: 'Parent', class_child_name: 'Child'}
        self.smart._origins = {v: k for k, v in self.smart._aliases.items()}
        expected = {
            'cls': {'class': 'Parent', 'arg1': 'str???', 'arg2': 5},
            'smart_cls': {'class': 'Parent:Smart', 'arg1': 'str???', 'arg2': 5},
            'smart': {'class': 'Smart'},
            'unknown': '???',
            'smart_cls_with_default': {
                'class': 'Child:Smart',
                'arg1': 'str???',
                'arg2': 5,
            },
            'smart_cls_with_only_default_generic': {
                'class': 'Smart',
            },
        }

        actual = self.smart._representation(ClassComposition, skip_default=False)

        self.assertEqual(expected, actual)


class TestCheckCase(UnitCase):
    def setUp(self) -> None:
        self.smart = Smart(Class, arg1='str???', arg2=15)

    @patch('smartparams.smart.warnings')
    def test_check_and_init_class__check_false(self, warnings: Mock) -> None:
        self.smart.check_missings = False
        self.smart.check_overrides = False
        self.smart.check_typings = False

        obj = self.smart(arg2='88')

        self.assertIsInstance(obj, Class)
        self.assertEqual('88', obj.arg2)
        warnings.warn.assert_not_called()

    @patch('smartparams.smart.warnings')
    def test_check_and_init_class__check_missings_true(self, warnings: Mock) -> None:
        self.smart.check_missings = True
        self.smart.check_overrides = False
        self.smart.check_typings = False

        self.assertRaises(ValueError, self.smart, arg2='88')
        warnings.warn.assert_not_called()

    @patch('smartparams.smart.warnings')
    def test_check_and_init_class__check_missings_none(self, warnings: Mock) -> None:
        self.smart.check_missings = None
        self.smart.check_overrides = False
        self.smart.check_typings = False

        obj = self.smart(arg2='88')

        self.assertIsInstance(obj, Class)
        self.assertEqual('88', obj.arg2)
        warnings.warn.assert_called()

    @patch('smartparams.smart.warnings')
    def test_check_and_init_class__check_overrides_true(self, warnings: Mock) -> None:
        self.smart.check_missings = False
        self.smart.check_overrides = True
        self.smart.check_typings = False

        self.assertRaises(ValueError, self.smart, arg2='88')
        warnings.warn.assert_not_called()

    @patch('smartparams.smart.warnings')
    def test_check_and_init_class__check_overrides_none(self, warnings: Mock) -> None:
        self.smart.check_missings = False
        self.smart.check_overrides = None
        self.smart.check_typings = False

        obj = self.smart(arg2='88')

        self.assertIsInstance(obj, Class)
        self.assertEqual('88', obj.arg2)
        warnings.warn.assert_called()

    @patch('smartparams.smart.warnings')
    def test_check_and_init_class__check_typings_true(self, warnings: Mock) -> None:
        self.smart.check_missings = False
        self.smart.check_overrides = False
        self.smart.check_typings = True

        self.assertRaises(TypeError, self.smart, arg2='88')
        warnings.warn.assert_not_called()

    @patch('smartparams.smart.warnings')
    def test_check_and_init_class__check_typings_none(self, warnings: Mock) -> None:
        self.smart.check_missings = False
        self.smart.check_overrides = False
        self.smart.check_typings = None

        obj = self.smart(arg2='88')

        self.assertIsInstance(obj, Class)
        self.assertEqual('88', obj.arg2)
        warnings.warn.assert_called()

    def test_check_and_init_class__location(self) -> None:
        smart = Smart(Smart, nested=[{'class': 'Smart'}])
        smart._location = 'location'

        obj = smart()

        self.assertIsInstance(obj, Smart)
        self.assertEqual('location.nested.0', obj.get('nested')[0]._location)

    def test_check_and_init_class__raise(self) -> None:
        smart = Smart(RaiseClass)

        self.assertRaises(Exception, smart)



import inspect
from typing import Any
from unittest.mock import Mock

from smartparams import Smart
from smartparams.utils import get_hints, get_name, get_nested
from tests.custom_classes import Class, ClassCompositionChild
from tests.unit import UnitCase


class TestGetNested(UnitCase):
    def setUp(self) -> None:
        self.dict = dict(arg1='arg1', arg2=['arg2'], arg3={'arg31': 'a31', 'arg32': 'a32'})

    def test_nested(self) -> None:
        name = 'arg3.arg31'

        dictionary, key = get_nested(dictionary=self.dict, name=name)

        self.assertEqual('arg31', key)
        self.assertTupleEqual((('arg31', 'a31'), ('arg32', 'a32')), tuple(dictionary.items()))

    def test_nested__not_in_dictionary(self) -> None:
        name = 'missing.any'

        self.assertRaises(KeyError, get_nested, dictionary=self.dict, name=name)

    def test_nested__ensure_key_exists(self) -> None:
        name = 'arg3.missing'

        self.assertRaises(
            KeyError,
            get_nested,
            dictionary=self.dict,
            name=name,
            ensure_key_exists=True,
        )

    def test_nested__not_is_dictionary(self) -> None:
        name = 'arg3.arg31.a31'

        self.assertRaises(ValueError, get_nested, dictionary=self.dict, name=name)

    def test_nested__create_missing_directories(self) -> None:
        name = 'arg3.missing.key'

        dictionary, key = get_nested(
            dictionary=self.dict,
            name=name,
            create_missing_directories=True,
        )

        self.assertIsInstance(dictionary, dict)
        self.assertFalse(bool(dictionary))
        self.assertEqual('key', key)


class TestGetName(UnitCase):
    def test_get_name(self) -> None:
        test_cases = (
            (Mock, 'Mock'),
            (Mock(), 'Mock'),
            (123, 'int'),
            (type, 'type'),
            (None, 'NoneType'),
            (lambda: ..., '<lambda>'),
            ((i for i in range(1)), '<genexpr>'),
        )

        for cls, expected in test_cases:
            with self.subTest(expected=expected):
                actual = get_name(cls)  # type: ignore

                self.assertEqual(expected, actual)


class TestGetHints(UnitCase):
    def test_get_name(self) -> None:
        test_cases = (
            (
                ClassCompositionChild,
                {
                    'cls': Class,
                    'smart_cls': Smart[Class],
                    'smart': Smart,
                    'unknown': Any,
                    'no_type': Any,
                    'args': tuple[Any, ...],
                    'smart_cls_with_default': Smart[Class],
                    'smart_cls_with_only_default_generic': Smart,
                    'with_only_default_primitive': int,
                    'kwargs': dict[str, Any],
                },
            ),
        )

        for cls, expected in test_cases:
            signature = inspect.signature(cls)

            with self.subTest(cls=cls.__name__):
                actual = get_hints(signature)

                self.assertEqual(expected, actual)



from typing import Any

from smartparams.smart import Smart


class Class:
    def __init__(self, arg1: str, arg2: int = 5) -> None:
        self.arg1 = arg1
        self.arg2 = arg2


class ClassChild(Class):
    pass


class ClassComposition:
    def __init__(  # type: ignore
        self,
        cls: Class,
        smart_cls: Smart[Class],
        smart: Smart,
        unknown: Any,
        smart_cls_with_default: Smart[Class] = Smart(ClassChild, arg1="argument1"),
        smart_cls_with_only_default_generic=Smart(),
    ) -> None:
        self.cls = cls
        self.smart_cls = smart_cls
        self.smart = smart
        self.unknown = unknown
        self.smart_cls_with_default = smart_cls_with_default
        self.smart_cls_with_only_default_generic = smart_cls_with_only_default_generic


class ClassCompositionChild(ClassComposition):
    def __init__(  # type: ignore
        self,
        cls: Class,
        smart_cls: Smart[Class],
        smart: Smart,
        unknown: Any,
        no_type,
        *args: Any,
        smart_cls_with_default: Smart[Class] = Smart(ClassChild, arg1="argument1"),
        smart_cls_with_only_default_generic=Smart(),
        with_only_default_primitive=11,
        **kwargs: Any,
    ) -> None:
        super(ClassCompositionChild, self).__init__(
            cls=cls,
            smart_cls=smart_cls,
            smart=smart,
            unknown=unknown,
            smart_cls_with_default=smart_cls_with_default,
            smart_cls_with_only_default_generic=smart_cls_with_only_default_generic,
        )
        self.no_type = no_type
        self.args = args
        self.with_only_default_primitive = with_only_default_primitive
        self.kwargs = kwargs


class RaiseClass:
    def __init__(self) -> None:
        raise RuntimeError


def some_function() -> None:
    pass



import inspect
from pydoc import locate
from typing import Any, get_origin


def get_nested(
    dictionary: dict[str, Any],
    name: str,
    create_missing_directories: bool = False,
    ensure_key_exists: bool = False,
    separator: str = '.',
) -> tuple[dict[str, Any], str]:
    *nested_keys, last_key = name.split(separator)

    key_list = list()
    for key in nested_keys:
        key_list.append(key)
        if key not in dictionary:
            if create_missing_directories:
                dictionary[key] = dict()
            else:
                raise KeyError(f"Param '{separator.join(key_list)}' is not in dictionary.")

        dictionary = dictionary[key]
        if not isinstance(dictionary, dict):
            raise ValueError(f"Param '{separator.join(key_list)}' is not dictionary.")

    if ensure_key_exists and last_key not in dictionary:
        raise KeyError(f"Param '{last_key}' is not in dictionary.")

    return dictionary, last_key


def import_class(name: str) -> type:
    cls = locate(name)
    if cls is None:
        raise ValueError(f"Class '{name}' does not exist.")
    if not isinstance(cls, type):
        raise ValueError(f"Object '{name}' is not a class.")
    return cls


def get_name(obj: Any) -> str:
    try:
        return obj.__name__
    except AttributeError:
        return obj.__class__.__name__


def get_hints(signature: inspect.Signature) -> dict[str, Any]:
    type_hints: dict[str, Any] = {}
    for name, param in signature.parameters.items():
        if param.annotation is not inspect.Parameter.empty:
            param_type = param.annotation
        elif param.default is not inspect.Parameter.empty and param.default is not None:
            param_type = get_origin(param.default) or type(param.default)
        else:
            param_type = Any

        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            type_hints[name] = tuple[param_type, ...]  # type: ignore
        elif param.kind == inspect.Parameter.VAR_KEYWORD:
            type_hints[name] = dict[str, param_type]  # type: ignore
        else:
            type_hints[name] = param_type

    return type_hints



from pathlib import Path

from setuptools import find_packages, setup

setup(
    name='SmartParams',
    version='0.2.1',
    author='Mateusz Baran',
    author_email='mateusz.baran.sanok@gmail.com',
    maintainer='Mateusz Baran',
    maintainer_email='mateusz.baran.sanok@gmail.com',
    license='MIT',
    url='https://gitlab.com/mateusz.baran/smart-params',
    description='The tool for advanced project configuration with python object injection.',
    long_description=Path('README.md').read_text(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    python_requires='>=3.9, <4',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
    ],
    install_requires=[
        'PyYAML>=5.4.1',
        'typeguard>=2.13.3',
    ],
    extras_require=dict(
        dev=[
            'bump2version==1.0.1',
            'pytest==6.2.5',
            'pytest-black==0.3.12',
            'pytest-cov==3.0.0',
            'pytest-isort==2.0.0',
            'pytest-mypy==0.9.0',
        ],
    ),
)



import argparse
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Optional


@dataclass
class CLIArguments:
    path: Optional[Path]
    dump: bool
    save: bool
    params: list[str]


def parse_arguments(
    path: Optional[Path],
    dump: bool,
    save: bool,
) -> CLIArguments:
    return _parse_arguments(
        arguments=sys.argv[1:],
        path=path,
        dump=dump,
        save=save,
    )


def _parse_arguments(
    arguments: list[str],
    path: Optional[Path],
    dump: bool,
    save: bool,
) -> CLIArguments:
    parser = argparse.ArgumentParser()
    parser.add_argument('--path', type=Path, default=path)
    parser.add_argument('--dump', default=dump, action='store_true')
    parser.add_argument('--save', default=save, action='store_true')
    parser.add_argument('--set', '-s', default=list(), action='append')
    args = parser.parse_args(arguments)

    if path and args.path and not args.path.is_absolute():
        args.path = path.parent.joinpath(args.path)

    return CLIArguments(
        path=args.path,
        dump=args.dump,
        save=args.save,
        params=args.set,
    )



import os
import unittest

_INTEGRATION = bool(int(os.getenv('TEST_INTEGRATION', default='0')))


@unittest.skipUnless(_INTEGRATION, reason="Integration tests are disabled")
class IntegrationCase(unittest.TestCase):
    pass
